import re
from pathlib import Path

from phylogenie.treesimulator.tree import Tree


def parse_newick(newick: str, translations: dict[str, str] | None = None) -> Tree:
    newick = newick.strip()
    newick = re.sub(r"^\[\&[^\]]*\]", "", newick).strip()

    stack: list[list[Tree]] = []
    current_children: list[Tree] = []
    current_nodes: list[Tree] = []
    i = 0
    while True:

        def _read_chars(stoppers: list[str]) -> str:
            nonlocal i
            chars = ""
            while i < len(newick) and newick[i] not in stoppers:
                chars += newick[i]
                i += 1
            if i == len(newick):
                raise ValueError(f"Expected one of {stoppers}, got end of string")
            return chars

        if newick[i] == "(":
            stack.append(current_nodes)
            current_nodes = []
            i += 1
            continue

        name = _read_chars([":", "[", ",", ")", ";"])
        if translations is not None and name in translations:
            name = translations[name]
        current_node = Tree(name)

        if newick[i] == "[":
            i += 1
            if newick[i] != "&":
                raise ValueError("Expected '[&' at the start of node features")
            i += 1
            features = re.split(r",(?=[^,]+=)", _read_chars(["]"]))
            i += 1
            for feature in features:
                key, value = feature.split("=")
                try:
                    current_node.set(key, eval(value))
                except Exception:
                    current_node.set(key, value)

        if newick[i] == ":":
            i += 1
            current_node.branch_length = float(_read_chars([",", ")", ";"]))

        for node in current_children:
            current_node.add_child(node)
            current_children = []
        current_nodes.append(current_node)

        if newick[i] == ")":
            current_children = current_nodes
            current_nodes = stack.pop()
        elif newick[i] == ";":
            return current_node

        i += 1


def load_newick(filepath: str | Path) -> Tree | list[Tree]:
    with open(filepath, "r") as file:
        trees = [parse_newick(newick) for newick in file]
    return trees[0] if len(trees) == 1 else trees


def to_newick(tree: Tree) -> str:
    children_newick = ",".join([to_newick(child) for child in tree.children])
    newick = tree.name
    if tree.metadata:
        reprs = {k: repr(v).replace("'", '"') for k, v in tree.metadata.items()}
        for k, r in reprs.items():
            if "," in k or "=" in k or "]" in k:
                raise ValueError(
                    f"Invalid feature key `{k}`: keys must not contain ',', '=', or ']'"
                )
            if "=" in r or "]" in r:
                raise ValueError(
                    f"Invalid value  `{r}` for feature `{k}`: values must not contain '=' or ']'"
                )
        features = [f"{k}={repr}" for k, repr in reprs.items()]
        newick += f"[&{','.join(features)}]"
    if children_newick:
        newick = f"({children_newick}){newick}"
    if tree.branch_length is not None:
        newick += f":{tree.branch_length}"
    return newick


def dump_newick(trees: Tree | list[Tree], filepath: str | Path) -> None:
    if isinstance(trees, Tree):
        trees = [trees]
    with open(filepath, "w") as file:
        for t in trees:
            file.write(to_newick(t) + ";\n")



from abc import abstractmethod
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Callable, Literal

import numpy as np
from numpy.random import default_rng
from pydantic import Field

import phylogenie.generators.configs as cfg
from phylogenie.generators.configs import Distribution
from phylogenie.generators.dataset import DatasetGenerator, DataType
from phylogenie.generators.factories import (
    data,
    integer,
    mutations,
    scalar,
    skyline_matrix,
    skyline_parameter,
    skyline_vector,
)
from phylogenie.treesimulator import (
    Event,
    EventType,
    Feature,
    Tree,
    dump_newick,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_contact_tracing_events,
    get_epidemiological_events,
    get_FBD_events,
    set_features,
    simulate_tree,
)


class ParameterizationType(str, Enum):
    CANONICAL = "canonical"
    EPIDEMIOLOGICAL = "epidemiological"
    FBD = "FBD"
    BD = "BD"
    BDEI = "BDEI"
    BDSS = "BDSS"


class TreeDatasetGenerator(DatasetGenerator):
    data_type: Literal[DataType.TREES] = DataType.TREES
    mutations: list[cfg.Mutation] = Field(default_factory=lambda: [])
    rates_to_log: list[EventType] | None = None
    n_tips: cfg.Integer | None = None
    max_time: cfg.Scalar = np.inf
    init_state: str | None = None
    sampling_probability_at_present: cfg.Scalar = 0.0
    timeout: float = np.inf
    node_features: list[Feature] | None = None
    acceptance_criterion: str | None = None

    @abstractmethod
    def _get_events(self, data: dict[str, Any]) -> list[Event]: ...

    def simulate_one(
        self, data: dict[str, Any], seed: int | None = None
    ) -> tuple[Tree, dict[str, Any]]:
        init_state = (
            self.init_state
            if self.init_state is None
            else self.init_state.format(**data)
        )
        events = self._get_events(data)
        states = {e.state for e in events}
        events += mutations(
            self.mutations, data, states, self.rates_to_log, default_rng(seed)
        )
        acceptance_criterion: None | Callable[[Tree], bool] = (
            None
            if self.acceptance_criterion is None
            else lambda tree: eval(
                self.acceptance_criterion, {}, {"tree": tree}  # pyright: ignore
            )
        )
        return simulate_tree(
            events=events,
            n_tips=None if self.n_tips is None else integer(self.n_tips, data),
            max_time=scalar(self.max_time, data),
            init_state=init_state,
            sampling_probability_at_present=scalar(
                self.sampling_probability_at_present, data
            ),
            seed=seed,
            timeout=self.timeout,
            acceptance_criterion=acceptance_criterion,
        )

    def generate_one(
        self,
        filename: str,
        context: dict[str, Distribution] | None = None,
        seed: int | None = None,
    ) -> dict[str, Any]:
        d = {"file_id": Path(filename).stem}
        rng = default_rng(seed)
        while True:
            try:
                d.update(data(context, rng))
                tree, metadata = self.simulate_one(d, seed)
                if self.node_features is not None:
                    set_features(tree, self.node_features)
                dump_newick(tree, f"{filename}.nwk")
                break
            except TimeoutError:
                print("Simulation timed out, retrying with different parameters...")
        return d | metadata


class CanonicalTreeDatasetGenerator(TreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.CANONICAL] = (
        ParameterizationType.CANONICAL
    )
    states: list[str]
    sampling_rates: cfg.SkylineVector = 0
    remove_after_sampling: bool = False
    birth_rates: cfg.SkylineVector = 0
    death_rates: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    birth_rates_among_states: cfg.SkylineMatrix = None

    def _get_events(self, data: dict[str, Any]) -> list[Event]:
        return get_canonical_events(
            states=self.states,
            sampling_rates=skyline_vector(self.sampling_rates, data),
            remove_after_sampling=self.remove_after_sampling,
            birth_rates=skyline_vector(self.birth_rates, data),
            death_rates=skyline_vector(self.death_rates, data),
            migration_rates=skyline_matrix(self.migration_rates, data),
            birth_rates_among_states=skyline_matrix(
                self.birth_rates_among_states, data
            ),
        )


class FBDTreeDatasetGenerator(TreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.FBD] = ParameterizationType.FBD
    states: list[str]
    sampling_proportions: cfg.SkylineVector = 0
    diversification: cfg.SkylineVector = 0
    turnover: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    diversification_between_states: cfg.SkylineMatrix = None

    def _get_events(self, data: dict[str, Any]) -> list[Event]:
        return get_FBD_events(
            states=self.states,
            diversification=skyline_vector(self.diversification, data),
            turnover=skyline_vector(self.turnover, data),
            sampling_proportions=skyline_vector(self.sampling_proportions, data),
            migration_rates=skyline_matrix(self.migration_rates, data),
            diversification_between_states=skyline_matrix(
                self.diversification_between_states, data
            ),
        )


class ContactTracingTreeDatasetGenerator(TreeDatasetGenerator):
    max_notified_contacts: cfg.Integer = 1
    notification_probability: cfg.SkylineParameter = 0.0
    sampling_rate_after_notification: cfg.SkylineParameter = 2**32
    samplable_states_after_notification: list[str] | None = None

    @abstractmethod
    def _get_base_events(self, data: dict[str, Any]) -> list[Event]: ...

    def _get_events(self, data: dict[str, Any]) -> list[Event]:
        events = self._get_base_events(data)
        if self.notification_probability:
            events = get_contact_tracing_events(
                events=events,
                max_notified_contacts=integer(self.max_notified_contacts, data),
                notification_probability=skyline_parameter(
                    self.notification_probability, data
                ),
                sampling_rate_after_notification=skyline_parameter(
                    self.sampling_rate_after_notification, data
                ),
                samplable_states_after_notification=self.samplable_states_after_notification,
            )
        return events


class EpidemiologicalTreeDatasetGenerator(ContactTracingTreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.EPIDEMIOLOGICAL] = (
        ParameterizationType.EPIDEMIOLOGICAL
    )
    states: list[str]
    sampling_proportions: cfg.SkylineVector
    reproduction_numbers: cfg.SkylineVector = 0
    become_uninfectious_rates: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    reproduction_numbers_among_states: cfg.SkylineMatrix = None

    def _get_base_events(self, data: dict[str, Any]) -> list[Event]:
        return get_epidemiological_events(
            states=self.states,
            reproduction_numbers=skyline_vector(self.reproduction_numbers, data),
            become_uninfectious_rates=skyline_vector(
                self.become_uninfectious_rates, data
            ),
            sampling_proportions=skyline_vector(self.sampling_proportions, data),
            migration_rates=skyline_matrix(self.migration_rates, data),
            reproduction_numbers_among_states=skyline_matrix(
                self.reproduction_numbers_among_states, data
            ),
        )


class BDTreeDatasetGenerator(ContactTracingTreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.BD] = ParameterizationType.BD
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_base_events(self, data: dict[str, Any]) -> list[Event]:
        return get_BD_events(
            reproduction_number=skyline_parameter(self.reproduction_number, data),
            infectious_period=skyline_parameter(self.infectious_period, data),
            sampling_proportion=skyline_parameter(self.sampling_proportion, data),
        )


class BDEITreeDatasetGenerator(ContactTracingTreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.BDEI] = ParameterizationType.BDEI
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    incubation_period: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_base_events(self, data: dict[str, Any]) -> list[Event]:
        return get_BDEI_events(
            reproduction_number=skyline_parameter(self.reproduction_number, data),
            infectious_period=skyline_parameter(self.infectious_period, data),
            incubation_period=skyline_parameter(self.incubation_period, data),
            sampling_proportion=skyline_parameter(self.sampling_proportion, data),
        )


class BDSSTreeDatasetGenerator(ContactTracingTreeDatasetGenerator):
    parameterization: Literal[ParameterizationType.BDSS] = ParameterizationType.BDSS
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    superspreading_ratio: cfg.SkylineParameter
    superspreaders_proportion: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_base_events(self, data: dict[str, Any]) -> list[Event]:
        return get_BDSS_events(
            reproduction_number=skyline_parameter(self.reproduction_number, data),
            infectious_period=skyline_parameter(self.infectious_period, data),
            superspreading_ratio=skyline_parameter(self.superspreading_ratio, data),
            superspreaders_proportion=skyline_parameter(
                self.superspreaders_proportion, data
            ),
            sampling_proportion=skyline_parameter(self.sampling_proportion, data),
        )


TreeDatasetGeneratorConfig = Annotated[
    CanonicalTreeDatasetGenerator
    | EpidemiologicalTreeDatasetGenerator
    | FBDTreeDatasetGenerator
    | BDTreeDatasetGenerator
    | BDEITreeDatasetGenerator
    | BDSSTreeDatasetGenerator,
    Field(discriminator="parameterization"),
]



import re
from copy import deepcopy
from typing import Any, Callable

from numpy.random import Generator

from phylogenie.skyline import SkylineParameterLike
from phylogenie.treesimulator.events.base import Event, EventType
from phylogenie.treesimulator.events.contact_tracing import (
    BirthWithContactTracing,
    SamplingWithContactTracing,
)
from phylogenie.treesimulator.events.core import Birth, Death, Migration, Sampling
from phylogenie.treesimulator.model import Model

MUTATION_PREFIX = "MUT-"
NEXT_MUTATION_ID = "NEXT_MUTATION_ID"


def _get_mutation(state: str) -> str | None:
    return state.split(".")[0] if state.startswith(MUTATION_PREFIX) else None


def _get_mutated_state(mutation_id: int, state: str) -> str:
    if state.startswith(MUTATION_PREFIX):
        _, state = state.split(".")
    return f"{MUTATION_PREFIX}{mutation_id}.{state}"


def get_mutation_id(node_name: str) -> int:
    match = re.search(rf"{MUTATION_PREFIX}(\d+)\.", node_name)
    if match:
        return int(match.group(1))
    return 0


class Mutation(Event):
    type = EventType.MUTATION

    def __init__(
        self,
        state: str,
        rate: SkylineParameterLike,
        rate_scalers: dict[EventType, Callable[[], float]],
        rates_to_log: list[EventType] | None = None,
    ):
        super().__init__(state, rate)
        self.rate_scalers = rate_scalers
        self.rates_to_log = [] if rates_to_log is None else rates_to_log

    def apply(
        self, model: Model, events: list[Event], time: float, rng: Generator
    ) -> dict[str, Any]:
        if NEXT_MUTATION_ID not in model.metadata:
            model[NEXT_MUTATION_ID] = 0
        model[NEXT_MUTATION_ID] += 1
        mutation_id = model[NEXT_MUTATION_ID]

        individual = self.draw_individual(model, rng)
        model.migrate(individual, _get_mutated_state(mutation_id, self.state), time)

        rate_scalers: dict[EventType, float] = {
            target_type: rate_scaler()
            for target_type, rate_scaler in self.rate_scalers.items()
        }

        metadata: dict[str, Any] = {}
        for event in [
            deepcopy(e)
            for e in events
            if _get_mutation(self.state) == _get_mutation(e.state)
        ]:
            event.state = _get_mutated_state(mutation_id, event.state)

            if isinstance(event, Birth | BirthWithContactTracing):
                event.child_state = _get_mutated_state(mutation_id, event.child_state)
                metadata_key = f"birth_rate_from_{event.state}_to_{event.child_state}"
            elif isinstance(event, Migration):
                event.target_state = _get_mutated_state(mutation_id, event.target_state)
                metadata_key = (
                    f"migration_rate_from_{event.state}_to_{event.target_state}"
                )
            elif isinstance(
                event, Mutation | Death | Sampling | SamplingWithContactTracing
            ):
                metadata_key = f"{event.type}_rate_for_{event.state}"
            else:
                raise ValueError(
                    f"Mutation not implemented for event of type {type(event)}."
                )

            event.rate *= rate_scalers.get(event.type, 1)
            if event.type in self.rates_to_log:
                metadata[metadata_key] = (
                    event.rate.value[0]
                    if len(event.rate.value) == 1
                    else list(event.rate.value)
                )

            events.append(event)

        return metadata

    def __repr__(self) -> str:
        return f"Mutation(state={self.state}, rate={self.rate}, rate_scalers={self.rate_scalers})"



from phylogenie.treesimulator.events import (
    Birth,
    BirthWithContactTracing,
    Death,
    Event,
    EventType,
    Migration,
    Mutation,
    Sampling,
    SamplingWithContactTracing,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_contact_tracing_events,
    get_epidemiological_events,
    get_FBD_events,
    get_mutation_id,
)
from phylogenie.treesimulator.features import Feature, set_features
from phylogenie.treesimulator.gillespie import generate_trees, simulate_tree
from phylogenie.treesimulator.io import dump_newick, load_newick, load_nexus
from phylogenie.treesimulator.model import get_node_state
from phylogenie.treesimulator.tree import Tree
from phylogenie.treesimulator.utils import (
    compute_mean_leaf_pairwise_distance,
    compute_sackin_index,
    get_distance,
    get_mrca,
    get_node_depth_levels,
    get_node_depths,
    get_node_height_levels,
    get_node_heights,
    get_node_leaf_counts,
)

__all__ = [
    "Birth",
    "BirthWithContactTracing",
    "Death",
    "Event",
    "EventType",
    "Migration",
    "Mutation",
    "Sampling",
    "SamplingWithContactTracing",
    "get_BD_events",
    "get_BDEI_events",
    "get_BDSS_events",
    "get_canonical_events",
    "get_contact_tracing_events",
    "get_epidemiological_events",
    "get_FBD_events",
    "get_mutation_id",
    "Feature",
    "set_features",
    "simulate_tree",
    "dump_newick",
    "load_newick",
    "load_nexus",
    "generate_trees",
    "get_node_state",
    "Tree",
    "compute_mean_leaf_pairwise_distance",
    "compute_sackin_index",
    "get_distance",
    "get_mrca",
    "get_node_depth_levels",
    "get_node_depths",
    "get_node_height_levels",
    "get_node_heights",
    "get_node_leaf_counts",
]



from collections.abc import Iterable
from enum import Enum

from phylogenie.treesimulator.events.mutations import get_mutation_id
from phylogenie.treesimulator.model import get_node_state
from phylogenie.treesimulator.tree import Tree
from phylogenie.treesimulator.utils import (
    get_node_depth_levels,
    get_node_depths,
    get_node_height_levels,
    get_node_heights,
    get_node_leaf_counts,
)


def _get_states(tree: Tree) -> dict[Tree, str]:
    return {node: get_node_state(node.name) for node in tree}


def _get_mutations(tree: Tree) -> dict[Tree, int]:
    return {node: get_mutation_id(node.name) for node in tree}


class Feature(str, Enum):
    DEPTH = "depth"
    DEPTH_LEVEL = "depth_level"
    HEIGHT = "height"
    HEIGHT_LEVEL = "height_level"
    MUTATION = "mutation"
    N_LEAVES = "n_leaves"
    STATE = "state"


FEATURES_EXTRACTORS = {
    Feature.DEPTH: get_node_depths,
    Feature.DEPTH_LEVEL: get_node_depth_levels,
    Feature.HEIGHT: get_node_heights,
    Feature.HEIGHT_LEVEL: get_node_height_levels,
    Feature.MUTATION: _get_mutations,
    Feature.N_LEAVES: get_node_leaf_counts,
    Feature.STATE: _get_states,
}


def set_features(tree: Tree, features: Iterable[Feature]) -> None:
    for feature in features:
        feature_maps = FEATURES_EXTRACTORS[feature](tree)
        for node in tree:
            node[feature.value] = feature_maps[node]



from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy

from numpy.random import Generator

from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.treesimulator.events.base import Event, EventType
from phylogenie.treesimulator.events.core import Birth, Death, Migration, Sampling
from phylogenie.treesimulator.model import Model

CT_POSTFIX = "-CT"
CONTACTS_KEY = "CONTACTS"


def get_CT_state(state: str) -> str:
    return f"{state}{CT_POSTFIX}"


def is_CT_state(state: str) -> bool:
    return state.endswith(CT_POSTFIX)


class BirthWithContactTracing(Event):
    type = EventType.BIRTH

    def __init__(self, state: str, rate: SkylineParameterLike, child_state: str):
        super().__init__(state, rate)
        self.child_state = child_state

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        new_individual = model.birth_from(individual, self.child_state, time)
        if CONTACTS_KEY not in model.metadata:
            model[CONTACTS_KEY] = defaultdict(list)
        model[CONTACTS_KEY][individual].append(new_individual)
        model[CONTACTS_KEY][new_individual].append(individual)

    def __repr__(self) -> str:
        return f"BirthWithContactTracing(state={self.state}, rate={self.rate}, child_state={self.child_state})"


class SamplingWithContactTracing(Event):
    type = EventType.SAMPLING

    def __init__(
        self,
        state: str,
        rate: SkylineParameterLike,
        max_notified_contacts: int,
        notification_probability: SkylineParameterLike,
    ):
        super().__init__(state, rate)
        self.max_notified_contacts = max_notified_contacts
        self.notification_probability = skyline_parameter(notification_probability)

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        model.sample(individual, time, True)
        population = model.get_population()
        if CONTACTS_KEY not in model.metadata:
            return
        contacts = model[CONTACTS_KEY][individual]
        for contact in contacts[-self.max_notified_contacts :]:
            if contact in population:
                state = model.get_state(contact)
                p = self.notification_probability.get_value_at_time(time)
                if not is_CT_state(state) and rng.random() < p:
                    model.migrate(contact, get_CT_state(state), time)

    def __repr__(self) -> str:
        return f"SamplingWithContactTracing(state={self.state}, rate={self.rate}, max_notified_contacts={self.max_notified_contacts}, notification_probability={self.notification_probability})"


def get_contact_tracing_events(
    events: Sequence[Event],
    max_notified_contacts: int = 1,
    notification_probability: SkylineParameterLike = 0.0,
    sampling_rate_after_notification: SkylineParameterLike = 2**32,
    samplable_states_after_notification: list[str] | None = None,
) -> list[Event]:
    ct_events: list[Event] = []
    notification_probability = skyline_parameter(notification_probability)
    sampling_rate_after_notification = skyline_parameter(
        sampling_rate_after_notification
    )
    for event in [deepcopy(e) for e in events]:
        if isinstance(event, Migration):
            ct_events.append(event)
            ct_events.append(
                Migration(
                    get_CT_state(event.state),
                    event.rate,
                    get_CT_state(event.target_state),
                )
            )
        elif isinstance(event, Birth):
            ct_events.append(
                BirthWithContactTracing(event.state, event.rate, event.child_state)
            )
            ct_events.append(
                BirthWithContactTracing(
                    get_CT_state(event.state), event.rate, event.child_state
                )
            )
        elif isinstance(event, Sampling):
            if not event.removal:
                raise ValueError(
                    "Contact tracing requires removal to be set for all sampling events."
                )
            ct_events.append(
                SamplingWithContactTracing(
                    event.state,
                    event.rate,
                    max_notified_contacts,
                    notification_probability,
                )
            )
        elif isinstance(event, Death):
            ct_events.append(event)
        else:
            raise NotImplementedError(
                f"Unsupported event type {type(event)} for contact tracing."
            )

    for state in (
        samplable_states_after_notification
        if samplable_states_after_notification is not None
        else {e.state for e in events}
    ):
        ct_events.append(
            SamplingWithContactTracing(
                get_CT_state(state),
                sampling_rate_after_notification,
                max_notified_contacts,
                notification_probability,
            )
        )

    return ct_events



from types import MappingProxyType
from typing import Any, Mapping, Optional


class MetadataMixin:
    """A mixin that provides metadata management with dictionary-like access."""

    def __init__(self) -> None:
        self._metadata: dict[str, Any] = {}

    @property
    def metadata(self) -> Mapping[str, Any]:
        """Return a read-only view of all metadata."""
        return MappingProxyType(self._metadata)

    def set(self, key: str, value: Any) -> None:
        """Set or update a metadata value."""
        self._metadata[key] = value

    def update(self, metadata: Mapping[str, Any]) -> None:
        """Bulk update metadata values."""
        self._metadata.update(metadata)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """Get a metadata value, returning `default` if not found."""
        return self._metadata.get(key, default)

    def delete(self, key: str) -> None:
        """Delete a metadata if it exists, else do nothing."""
        self._metadata.pop(key, None)

    def clear(self) -> None:
        """Remove all metadata."""
        self._metadata.clear()

    # Dict-like behavior
    def __getitem__(self, key: str) -> Any:
        return self._metadata[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._metadata[key] = value



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

from numpy.random import Generator

from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.treesimulator.model import Model


class EventType(str, Enum):
    BIRTH = "birth"
    DEATH = "death"
    MIGRATION = "migration"
    SAMPLING = "sampling"
    MUTATION = "mutation"


class Event(ABC):
    type: EventType

    def __init__(self, state: str, rate: SkylineParameterLike):
        self.state = state
        self.rate = skyline_parameter(rate)
        if any(v < 0 for v in self.rate.value):
            raise ValueError("Event rates must be non-negative.")

    def draw_individual(self, model: Model, rng: Generator) -> int:
        return rng.choice(model.get_population(self.state))

    def get_propensity(self, model: Model, time: float) -> float:
        n_individuals = model.count_individuals(self.state)
        rate = self.rate.get_value_at_time(time)
        return rate * n_individuals

    @abstractmethod
    def apply(
        self, model: Model, events: "list[Event]", time: float, rng: Generator
    ) -> dict[str, Any] | None: ...



import math
from math import comb

from phylogenie.treesimulator.tree import Tree


def get_node_leaf_counts(tree: Tree) -> dict[Tree, int]:
    n_leaves: dict[Tree, int] = {}
    for node in tree.postorder_traversal():
        n_leaves[node] = sum(n_leaves[child] for child in node.children) or 1
    return n_leaves


def get_node_depth_levels(tree: Tree) -> dict[Tree, int]:
    depth_levels: dict[Tree, int] = {tree: tree.depth_level}
    for node in tree.iter_descendants():
        depth_levels[node] = depth_levels[node.parent] + 1  # pyright: ignore
    return depth_levels


def get_node_depths(tree: Tree) -> dict[Tree, float]:
    depths: dict[Tree, float] = {tree: tree.depth}
    for node in tree.iter_descendants():
        parent_depth = depths[node.parent]  # pyright: ignore
        depths[node] = node.branch_length_or_raise() + parent_depth
    return depths


def get_node_height_levels(tree: Tree) -> dict[Tree, int]:
    height_levels: dict[Tree, int] = {}
    for node in tree.postorder_traversal():
        height_levels[node] = (
            0
            if node.is_leaf()
            else max(1 + height_levels[child] for child in node.children)
        )
    return height_levels


def get_node_heights(tree: Tree) -> dict[Tree, float]:
    heights: dict[Tree, float] = {}
    for node in tree.postorder_traversal():
        heights[node] = (
            0
            if node.is_leaf()
            else max(
                child.branch_length_or_raise() + heights[child]
                for child in node.children
            )
        )
    return heights


def get_mrca(node1: Tree, node2: Tree) -> Tree:
    node1_ancestors = set(node1.iter_upward())
    for node2_ancestor in node2.iter_upward():
        if node2_ancestor in node1_ancestors:
            return node2_ancestor
    raise ValueError(f"No common ancestor found between node {node1} and node {node2}.")


def get_path(node1: Tree, node2: Tree) -> list[Tree]:
    mrca = get_mrca(node1, node2)
    return [
        *node1.iter_upward(stop=mrca.parent),
        *reversed(list(node2.iter_upward(stop=mrca))),
    ]


def count_hops(node1: Tree, node2: Tree) -> int:
    return len(get_path(node1, node2)) - 1


def get_distance(node1: Tree, node2: Tree) -> float:
    mrca = get_mrca(node1, node2)
    path = get_path(node1, node2)
    path.remove(mrca)
    return sum(node.branch_length_or_raise() for node in path)


def compute_sackin_index(tree: Tree, normalize: bool = False) -> float:
    depth_levels = get_node_depth_levels(tree)
    sackin_index = sum(dl for node, dl in depth_levels.items() if node.is_leaf())
    if normalize:
        if not tree.is_binary():
            raise ValueError(
                "Normalized Sackin index is only defined for binary trees."
            )
        n = tree.n_leaves
        h = math.floor(math.log2(n))
        min_sackin_index = n * (h + 2) - 2 ** (h + 1)
        max_sackin_index = n * (n - 1) / 2
        return (sackin_index - min_sackin_index) / (max_sackin_index - min_sackin_index)
    return sackin_index


def compute_mean_leaf_pairwise_distance(tree: Tree) -> float:
    leaves = tree.get_leaves()
    n_leaves = len(leaves)
    if n_leaves < 2:
        return 0.0

    total_distance = sum(
        get_distance(leaves[i], leaves[j])
        for i in range(n_leaves)
        for j in range(i + 1, n_leaves)
    )
    return total_distance / comb(n_leaves, 2)



import re
from collections.abc import Iterator
from pathlib import Path

from phylogenie.treesimulator.io.newick import parse_newick
from phylogenie.treesimulator.tree import Tree


def _parse_translate_block(lines: Iterator[str]) -> dict[str, str]:
    translations: dict[str, str] = {}
    for line in lines:
        line = line.strip()
        match = re.match(r"(\d+)\s+['\"]?([^'\",;]+)['\"]?", line)
        if match is None:
            if ";" in line:
                return translations
            else:
                raise ValueError(f"Invalid translate line: {line}")
        translations[match.group(1)] = match.group(2)
    raise ValueError("Translate block not terminated with ';'")


def _parse_trees_block(lines: Iterator[str]) -> dict[str, Tree]:
    trees: dict[str, Tree] = {}
    translations = {}
    for line in lines:
        line = line.strip()
        if line.upper() == "TRANSLATE":
            translations = _parse_translate_block(lines)
        elif line.upper() == "END;":
            return trees
        else:
            match = re.match(r"^TREE\s*\*?\s+(\S+)\s*=\s*(.+)$", line, re.IGNORECASE)
            if match is None:
                raise ValueError(
                    f"Invalid tree line. Expected 'TREE <name> = <newick>'"
                )
            trees[match.group(1)] = parse_newick(match.group(2), translations)
    raise ValueError("Unterminated TREES block.")


def load_nexus(nexus_file: str | Path) -> dict[str, Tree]:
    with open(nexus_file, "r") as f:
        for line in f:
            if line.strip().upper() == "BEGIN TREES;":
                return _parse_trees_block(f)
    raise ValueError("No TREES block found in the NEXUS file.")



from collections import defaultdict
from dataclasses import dataclass

from phylogenie.mixins import MetadataMixin
from phylogenie.treesimulator.tree import Tree


@dataclass
class Individual:
    id: int
    node: Tree
    state: str


def _get_node_name(node_id: int, state: str) -> str:
    return f"{node_id}|{state}"


def get_node_state(node_name: str) -> str:
    if "|" not in node_name:
        raise ValueError(
            f"Invalid node name: {node_name} (expected format 'id|state')."
        )
    return node_name.split("|")[-1]


class Model(MetadataMixin):
    def __init__(self, init_state: str):
        super().__init__()
        self._next_node_id = 0
        self._next_individual_id = 0
        self._population: dict[int, Individual] = {}
        self._states: dict[str, set[int]] = defaultdict(set)
        self._sampled: set[str] = set()
        self._tree = self._get_new_individual(init_state).node

    @property
    def n_sampled(self) -> int:
        return len(self._sampled)

    def _get_new_node(self, state: str) -> Tree:
        self._next_node_id += 1
        node = Tree(_get_node_name(self._next_node_id, state))
        return node

    def _get_new_individual(self, state: str) -> Individual:
        self._next_individual_id += 1
        individual = Individual(
            self._next_individual_id, self._get_new_node(state), state
        )
        self._population[individual.id] = individual
        self._states[state].add(individual.id)
        return individual

    def _set_branch_length(self, node: Tree, time: float) -> None:
        if node.branch_length is not None:
            raise ValueError(f"Branch length of node {node.name} is already set.")
        node.branch_length = time if node.parent is None else time - node.parent.depth

    def _stem(self, individual: Individual, time: float) -> None:
        self._set_branch_length(individual.node, time)
        stem_node = self._get_new_node(individual.state)
        individual.node.add_child(stem_node)
        individual.node = stem_node

    def remove(self, id: int, time: float) -> None:
        individual = self._population[id]
        self._set_branch_length(individual.node, time)
        self._population.pop(id)
        self._states[individual.state].remove(id)

    def migrate(self, id: int, state: str, time: float) -> None:
        individual = self._population[id]
        self._states[individual.state].remove(id)
        individual.state = state
        self._states[state].add(id)
        self._stem(individual, time)

    def birth_from(self, id: int, state: str, time: float) -> int:
        individual = self._population[id]
        new_individual = self._get_new_individual(state)
        individual.node.add_child(new_individual.node)
        self._stem(individual, time)
        return new_individual.id

    def sample(self, id: int, time: float, removal: bool) -> None:
        individual = self._population[id]
        if removal:
            self._sampled.add(individual.node.name)
            self.remove(id, time)
        else:
            sample_node = self._get_new_node(individual.state)
            sample_node.branch_length = 0.0
            self._sampled.add(sample_node.name)
            individual.node.add_child(sample_node)
            self._stem(individual, time)

    def get_state(self, id: int) -> str:
        return self._population[id].state

    def get_sampled_tree(self) -> Tree:
        tree = self._tree.copy()
        for node in list(tree.postorder_traversal()):
            if node.name not in self._sampled and not node.children:
                if node.parent is None:
                    raise ValueError("No samples in the tree.")
                else:
                    node.parent.remove_child(node)
            elif len(node.children) == 1:
                (child,) = node.children
                child.set_parent(node.parent)
                child.branch_length += node.branch_length  # pyright: ignore
                if node.parent is None:
                    return child
                else:
                    node.parent.remove_child(node)
        return tree

    def get_full_tree(self) -> Tree:
        return self._tree.copy()

    def get_population(self, states: str | None = None) -> list[int]:
        if states is None:
            return list(self._population)
        return list(self._states[states])

    def count_individuals(self, states: str | None = None) -> int:
        if states is None:
            return len(self._population)
        return len(self._states[states])



from numpy.random import Generator

from phylogenie.skyline import (
    SkylineMatrixCoercible,
    SkylineParameterLike,
    SkylineVectorCoercible,
    skyline_matrix,
    skyline_vector,
)
from phylogenie.treesimulator.events.base import Event, EventType
from phylogenie.treesimulator.model import Model

INFECTIOUS_STATE = "I"
EXPOSED_STATE = "E"
SUPERSPREADER_STATE = "S"


class Birth(Event):
    type = EventType.BIRTH

    def __init__(self, state: str, rate: SkylineParameterLike, child_state: str):
        super().__init__(state, rate)
        self.child_state = child_state

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        model.birth_from(individual, self.child_state, time)

    def __repr__(self) -> str:
        return f"Birth(state={self.state}, rate={self.rate}, child_state={self.child_state})"


class Death(Event):
    type = EventType.DEATH

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        model.remove(individual, time)

    def __repr__(self) -> str:
        return f"Death(state={self.state}, rate={self.rate})"


class Migration(Event):
    type = EventType.MIGRATION

    def __init__(self, state: str, rate: SkylineParameterLike, target_state: str):
        super().__init__(state, rate)
        self.target_state = target_state

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        model.migrate(individual, self.target_state, time)

    def __repr__(self) -> str:
        return f"Migration(state={self.state}, rate={self.rate}, target_state={self.target_state})"


class Sampling(Event):
    type = EventType.SAMPLING

    def __init__(self, state: str, rate: SkylineParameterLike, removal: bool):
        super().__init__(state, rate)
        self.removal = removal

    def apply(self, model: Model, events: list[Event], time: float, rng: Generator):
        individual = self.draw_individual(model, rng)
        model.sample(individual, time, self.removal)

    def __repr__(self) -> str:
        return f"Sampling(state={self.state}, rate={self.rate}, removal={self.removal})"


def get_canonical_events(
    states: list[str],
    sampling_rates: SkylineVectorCoercible = 0,
    remove_after_sampling: bool = False,
    birth_rates: SkylineVectorCoercible = 0,
    death_rates: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible | None = None,
    birth_rates_among_states: SkylineMatrixCoercible | None = None,
) -> list[Event]:
    N = len(states)

    birth_rates = skyline_vector(birth_rates, N)
    death_rates = skyline_vector(death_rates, N)
    sampling_rates = skyline_vector(sampling_rates, N)

    events: list[Event] = []
    for i, state in enumerate(states):
        events.append(Birth(state, birth_rates[i], state))
        events.append(Death(state, death_rates[i]))
        events.append(Sampling(state, sampling_rates[i], remove_after_sampling))

    if migration_rates is not None:
        migration_rates = skyline_matrix(migration_rates, N, N - 1)
        for i, state in enumerate(states):
            for j, other_state in enumerate([s for s in states if s != state]):
                events.append(Migration(state, migration_rates[i, j], other_state))

    if birth_rates_among_states is not None:
        birth_rates_among_states = skyline_matrix(birth_rates_among_states, N, N - 1)
        for i, state in enumerate(states):
            for j, other_state in enumerate([s for s in states if s != state]):
                events.append(Birth(state, birth_rates_among_states[i, j], other_state))

    return [event for event in events if event.rate]


def get_FBD_events(
    states: list[str],
    sampling_proportions: SkylineVectorCoercible = 0,
    diversification: SkylineVectorCoercible = 0,
    turnover: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible | None = None,
    diversification_between_states: SkylineMatrixCoercible | None = None,
) -> list[Event]:
    N = len(states)

    diversification = skyline_vector(diversification, N)
    turnover = skyline_vector(turnover, N)
    sampling_proportions = skyline_vector(sampling_proportions, N)

    birth_rates = diversification / (1 - turnover)
    death_rates = turnover * birth_rates
    sampling_rates = sampling_proportions * death_rates
    birth_rates_among_states = (
        (skyline_matrix(diversification_between_states, N, N - 1) + death_rates)
        if diversification_between_states is not None
        else None
    )

    return get_canonical_events(
        states=states,
        sampling_rates=sampling_rates,
        remove_after_sampling=False,
        birth_rates=birth_rates,
        death_rates=death_rates,
        migration_rates=migration_rates,
        birth_rates_among_states=birth_rates_among_states,
    )


def get_epidemiological_events(
    states: list[str],
    sampling_proportions: SkylineVectorCoercible,
    reproduction_numbers: SkylineVectorCoercible = 0,
    become_uninfectious_rates: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible | None = None,
    reproduction_numbers_among_states: SkylineMatrixCoercible | None = None,
) -> list[Event]:
    N = len(states)

    reproduction_numbers = skyline_vector(reproduction_numbers, N)
    become_uninfectious_rates = skyline_vector(become_uninfectious_rates, N)
    sampling_proportions = skyline_vector(sampling_proportions, N)

    birth_rates = reproduction_numbers * become_uninfectious_rates
    sampling_rates = become_uninfectious_rates * sampling_proportions
    death_rates = become_uninfectious_rates - sampling_rates
    birth_rates_among_states = (
        (
            skyline_matrix(reproduction_numbers_among_states, N, N - 1)
            * become_uninfectious_rates
        )
        if reproduction_numbers_among_states is not None
        else None
    )

    return get_canonical_events(
        states=states,
        sampling_rates=sampling_rates,
        remove_after_sampling=True,
        birth_rates=birth_rates,
        death_rates=death_rates,
        migration_rates=migration_rates,
        birth_rates_among_states=birth_rates_among_states,
    )


def get_BD_events(
    reproduction_number: SkylineParameterLike,
    infectious_period: SkylineParameterLike,
    sampling_proportion: SkylineParameterLike,
) -> list[Event]:
    return get_epidemiological_events(
        states=[INFECTIOUS_STATE],
        reproduction_numbers=reproduction_number,
        become_uninfectious_rates=1 / infectious_period,
        sampling_proportions=sampling_proportion,
    )


def get_BDEI_events(
    reproduction_number: SkylineParameterLike,
    infectious_period: SkylineParameterLike,
    incubation_period: SkylineParameterLike,
    sampling_proportion: SkylineParameterLike,
) -> list[Event]:
    return get_epidemiological_events(
        states=[EXPOSED_STATE, INFECTIOUS_STATE],
        sampling_proportions=[0, sampling_proportion],
        become_uninfectious_rates=[0, 1 / infectious_period],
        reproduction_numbers_among_states=[[0], [reproduction_number]],
        migration_rates=[[1 / incubation_period], [0]],
    )


def get_BDSS_events(
    reproduction_number: SkylineParameterLike,
    infectious_period: SkylineParameterLike,
    superspreading_ratio: SkylineParameterLike,
    superspreaders_proportion: SkylineParameterLike,
    sampling_proportion: SkylineParameterLike,
) -> list[Event]:
    f_SS = superspreaders_proportion
    r_SS = superspreading_ratio
    R_0_IS = reproduction_number * f_SS / (1 + r_SS * f_SS - f_SS)
    R_0_SI = (reproduction_number - r_SS * R_0_IS) * r_SS
    R_0_S = r_SS * R_0_IS
    R_0_I = R_0_SI / r_SS
    return get_epidemiological_events(
        states=[INFECTIOUS_STATE, SUPERSPREADER_STATE],
        reproduction_numbers=[R_0_I, R_0_S],
        reproduction_numbers_among_states=[[R_0_IS], [R_0_SI]],
        become_uninfectious_rates=1 / infectious_period,
        sampling_proportions=sampling_proportion,
    )



from collections.abc import Iterator
from dataclasses import dataclass

import numpy as np


@dataclass
class Sequence:
    id: str
    chars: str
    time: float | None = None


class MSA:
    def __init__(self, sequences: list[Sequence]):
        self.sequences = sequences
        lengths = {len(sequence.chars) for sequence in sequences}
        if len(lengths) > 1:
            raise ValueError(
                f"All sequences in the alignment must have the same length (got lengths: {lengths})"
            )

    @property
    def ids(self) -> list[str]:
        return [sequence.id for sequence in self.sequences]

    @property
    def times(self) -> list[float]:
        times: list[float] = []
        for sequence in self:
            if sequence.time is None:
                raise ValueError(f"Time is not set for sequence {sequence.id}.")
            times.append(sequence.time)
        return times

    @property
    def alignment(self) -> list[list[str]]:
        return [list(sequence.chars) for sequence in self.sequences]

    @property
    def n_sequences(self) -> int:
        return len(self.sequences)

    @property
    def n_sites(self) -> int:
        return len(self.alignment[0])

    @property
    def shape(self) -> tuple[int, int]:
        return self.n_sequences, self.n_sites

    def __len__(self) -> int:
        return self.n_sequences

    def __getitem__(self, item: int) -> Sequence:
        return self.sequences[item]

    def __iter__(self) -> Iterator[Sequence]:
        return iter(self.sequences)

    def count_informative_sites(self) -> int:
        n_informative_sites = 0
        for column in np.array(self.alignment).T:
            column: np.typing.NDArray[np.str_]
            _, char_counts = np.unique(column, return_counts=True)
            is_informative_char = char_counts >= 2
            if (is_informative_char).sum() >= 2:
                n_informative_sites += 1
        return n_informative_sites

    def count_unique_sequences(self) -> int:
        return len(np.unique(self.alignment, axis=0))



import re
from typing import Any, Callable

import numpy as np
from numpy.random import Generator

import phylogenie.generators.configs as cfg
import phylogenie.generators.typeguards as ctg
import phylogenie.typeguards as tg
import phylogenie.typings as pgt
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
)
from phylogenie.treesimulator import EventType, Mutation


def _eval_expression(expression: str, data: dict[str, Any]) -> Any:
    return np.array(
        eval(
            expression,
            {"np": np, **{k: np.array(v) for k, v in data.items()}},
        )
    ).tolist()


def integer(x: cfg.Integer, data: dict[str, Any]) -> int:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if isinstance(e, int):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected an int."
        )
    return x


def scalar(x: cfg.Scalar, data: dict[str, Any]) -> pgt.Scalar:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if isinstance(e, pgt.Scalar):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar."
        )
    return x


def string(s: Any, data: dict[str, Any]) -> str:
    if not isinstance(s, str):
        return str(s)
    return re.sub(
        r"\{([^{}]+)\}", lambda match: str(_eval_expression(match.group(1), data)), s
    )  # Match content inside curly braces


def many_scalars(x: cfg.ManyScalars, data: dict[str, Any]) -> pgt.ManyScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a sequence of scalars."
        )
    return [scalar(v, data) for v in x]


def one_or_many_scalars(
    x: cfg.OneOrManyScalars, data: dict[str, Any]
) -> pgt.OneOrManyScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar or a sequence of them."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return many_scalars(x, data)


def skyline_parameter(
    x: cfg.SkylineParameter, data: dict[str, Any]
) -> SkylineParameterLike:
    if isinstance(x, cfg.Scalar):
        return scalar(x, data)
    return SkylineParameter(
        value=many_scalars(x.value, data),
        change_times=many_scalars(x.change_times, data),
    )


def skyline_vector(
    x: cfg.SkylineVector, data: dict[str, Any]
) -> SkylineVectorCoercible:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineVectorCoercible object (e.g., a scalar or a sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_many_skyline_parameter_configs(x):
        return [skyline_parameter(p, data) for p in x]

    assert isinstance(x, cfg.SkylineVectorModel)

    change_times = many_scalars(x.change_times, data)
    if isinstance(x.value, str):
        e = _eval_expression(x.value, data)
        if tg.is_many_one_or_many_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineVector (expected a sequence composed of scalars and/or sequences of scalars)."
            )
    else:
        value = [one_or_many_scalars(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        return SkylineParameter(value=value, change_times=change_times)

    Ns = {len(elem) for elem in value if tg.is_many(elem)}
    if len(Ns) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineVector config must be scalars or have the same length (config {x.value} yielded value={value} with inconsistent lengths {Ns})."
        )
    (N,) = Ns
    value = [[p] * N if isinstance(p, pgt.Scalar) else p for p in value]

    return SkylineVector(value=value, change_times=change_times)


def one_or_many_2D_scalars(
    x: cfg.OneOrMany2DScalars, data: dict[str, Any]
) -> pgt.OneOrMany2DScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a nested (2D) sequence of scalars."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return [many_scalars(v, data) for v in x]


def skyline_matrix(
    x: cfg.SkylineMatrix, data: dict[str, Any]
) -> SkylineMatrixCoercible | None:
    if x is None:
        return None

    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineMatrixCoercible object (e.g., a scalar or a nested (2D) sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_many_skyline_vector_configs(x):
        return [skyline_vector(v, data) for v in x]

    assert isinstance(x, cfg.SkylineMatrixModel)

    change_times = many_scalars(x.change_times, data)
    if isinstance(x.value, str):
        e = _eval_expression(x.value, data)
        if tg.is_many_one_or_many_2D_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineMatrix (expected a sequence composed of scalars and/or nested (2D) sequences of scalars)."
            )
    else:
        value = [one_or_many_2D_scalars(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        return SkylineParameter(value=value, change_times=change_times)

    shapes: set[tuple[int, int]] = set()
    for elem in value:
        if tg.is_many_2D_scalars(elem):
            Ms = len(elem)
            Ns = {len(row) for row in elem}
            if len(Ns) > 1:
                raise ValueError(
                    f"The values of a SkylineMatrix config must be scalars or nested (2D) lists of them with a consistent row length (config {x.value} yielded element {elem} with row lengths {Ns})."
                )
            shapes.add((Ms, Ns.pop()))

    if len(shapes) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineMatrix config must be scalars or nested (2D) lists of them with the same shape (config {x.value} yielded value={value} with inconsistent shapes {shapes})."
        )
    ((M, N),) = shapes
    value = [[[e] * N] * M if isinstance(e, pgt.Scalar) else e for e in value]

    return SkylineMatrix(value=value, change_times=change_times)


def distribution(x: cfg.Distribution, data: dict[str, Any]) -> cfg.Distribution:
    args = x.args
    for arg_name, arg_value in args.items():
        if isinstance(arg_value, str):
            args[arg_name] = _eval_expression(arg_value, data)
    return cfg.Distribution(type=x.type, **args)


def mutations(
    x: list[cfg.Mutation],
    data: dict[str, Any],
    states: set[str],
    rates_to_log: list[EventType] | None,
    rng: Generator,
) -> list[Mutation]:
    mutations: list[Mutation] = []
    for m in x:
        rate = skyline_parameter(m.rate, data)
        rate_scalers: dict[EventType, Callable[[], float]] = {
            k: lambda: distribution(v, data)(rng) for k, v in m.rate_scalers.items()
        }
        if m.state is None:
            mutations.extend(
                Mutation(s, rate, rate_scalers, rates_to_log) for s in states
            )
        else:
            mutations.append(Mutation(m.state, rate, rate_scalers, rates_to_log))
    return mutations


def data(context: dict[str, cfg.Distribution] | None, rng: Generator) -> dict[str, Any]:
    if context is None:
        return {}
    data: dict[str, Any] = {}
    for k, v in context.items():
        data[k] = np.array(distribution(v, data)(rng)).tolist()
    return data



from phylogenie.draw import Coloring, draw_tree
from phylogenie.generators import (
    AliSimDatasetGenerator,
    BDEITreeDatasetGenerator,
    BDSSTreeDatasetGenerator,
    BDTreeDatasetGenerator,
    CanonicalTreeDatasetGenerator,
    DatasetGenerator,
    DatasetGeneratorConfig,
    EpidemiologicalTreeDatasetGenerator,
    FBDTreeDatasetGenerator,
    TreeDatasetGeneratorConfig,
)
from phylogenie.io import load_fasta
from phylogenie.msa import MSA
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
    skyline_matrix,
    skyline_parameter,
    skyline_vector,
)
from phylogenie.treesimulator import (
    Birth,
    BirthWithContactTracing,
    Death,
    Event,
    EventType,
    Feature,
    Migration,
    Mutation,
    Sampling,
    SamplingWithContactTracing,
    Tree,
    compute_mean_leaf_pairwise_distance,
    compute_sackin_index,
    dump_newick,
    generate_trees,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_contact_tracing_events,
    get_distance,
    get_epidemiological_events,
    get_FBD_events,
    get_mrca,
    get_mutation_id,
    get_node_depth_levels,
    get_node_depths,
    get_node_height_levels,
    get_node_heights,
    get_node_leaf_counts,
    get_node_state,
    load_newick,
    load_nexus,
    set_features,
    simulate_tree,
)

__all__ = [
    "Coloring",
    "draw_tree",
    "AliSimDatasetGenerator",
    "BDEITreeDatasetGenerator",
    "BDSSTreeDatasetGenerator",
    "BDTreeDatasetGenerator",
    "CanonicalTreeDatasetGenerator",
    "DatasetGenerator",
    "DatasetGeneratorConfig",
    "EpidemiologicalTreeDatasetGenerator",
    "FBDTreeDatasetGenerator",
    "TreeDatasetGeneratorConfig",
    "load_fasta",
    "MSA",
    "SkylineMatrix",
    "SkylineMatrixCoercible",
    "SkylineParameter",
    "SkylineParameterLike",
    "SkylineVector",
    "SkylineVectorCoercible",
    "SkylineVectorLike",
    "skyline_matrix",
    "skyline_parameter",
    "skyline_vector",
    "Birth",
    "BirthWithContactTracing",
    "Death",
    "Event",
    "EventType",
    "Feature",
    "Migration",
    "Mutation",
    "Sampling",
    "SamplingWithContactTracing",
    "Tree",
    "compute_mean_leaf_pairwise_distance",
    "compute_sackin_index",
    "dump_newick",
    "generate_trees",
    "get_BD_events",
    "get_BDEI_events",
    "get_BDSS_events",
    "get_canonical_events",
    "get_contact_tracing_events",
    "get_distance",
    "get_epidemiological_events",
    "get_FBD_events",
    "get_mrca",
    "get_mutation_id",
    "get_node_depth_levels",
    "get_node_depths",
    "get_node_height_levels",
    "get_node_heights",
    "get_node_leaf_counts",
    "get_node_state",
    "load_newick",
    "load_nexus",
    "set_features",
    "simulate_tree",
]



from collections import deque
from collections.abc import Callable, Iterator
from typing import Any

from phylogenie.mixins import MetadataMixin


class Tree(MetadataMixin):
    def __init__(self, name: str = "", branch_length: float | None = None):
        super().__init__()
        self.name = name
        self.branch_length = branch_length
        self._parent: Tree | None = None
        self._children: list[Tree] = []

    # ----------------
    # Basic properties
    # ----------------
    # Properties related to parent-child relationships.

    @property
    def children(self) -> tuple["Tree", ...]:
        return tuple(self._children)

    @property
    def parent(self) -> "Tree | None":
        return self._parent

    def add_child(self, child: "Tree") -> "Tree":
        if child.parent is not None:
            raise ValueError(f"Node {child.name} already has a parent.")
        child._parent = self
        self._children.append(child)
        return self

    def remove_child(self, child: "Tree") -> None:
        self._children.remove(child)
        child._parent = None

    def set_parent(self, parent: "Tree | None"):
        if self.parent is not None:
            self.parent.remove_child(self)
        self._parent = parent
        if parent is not None:
            parent._children.append(self)

    def is_leaf(self) -> bool:
        return not self.children

    def get_leaves(self) -> tuple["Tree", ...]:
        return tuple(node for node in self if node.is_leaf())

    def is_internal(self) -> bool:
        return not self.is_leaf()

    def get_internal_nodes(self) -> tuple["Tree", ...]:
        return tuple(node for node in self if node.is_internal())

    def is_binary(self) -> bool:
        return all(len(node.children) in (0, 2) for node in self)

    # --------------
    # Tree traversal
    # --------------
    # Methods for traversing the tree in various orders.

    def iter_ancestors(self, stop: "Tree | None" = None) -> Iterator["Tree"]:
        node = self
        while True:
            if node.parent is None:
                if stop is None:
                    return
                raise ValueError("Reached root without encountering stop node.")
            node = node.parent
            if node == stop:
                return
            yield node

    def iter_upward(self, stop: "Tree | None" = None) -> Iterator["Tree"]:
        if self == stop:
            return
        yield self
        yield from self.iter_ancestors(stop=stop)

    def iter_descendants(self) -> Iterator["Tree"]:
        for child in self.children:
            yield child
            yield from child.iter_descendants()

    def preorder_traversal(self) -> Iterator["Tree"]:
        yield self
        yield from self.iter_descendants()

    def inorder_traversal(self) -> Iterator["Tree"]:
        if self.is_leaf():
            yield self
            return
        if len(self.children) != 2:
            raise ValueError("Inorder traversal is only defined for binary trees.")
        left, right = self.children
        yield from left.inorder_traversal()
        yield self
        yield from right.inorder_traversal()

    def postorder_traversal(self) -> Iterator["Tree"]:
        for child in self.children:
            yield from child.postorder_traversal()
        yield self

    def breadth_first_traversal(self) -> Iterator["Tree"]:
        queue: deque["Tree"] = deque([self])
        while queue:
            node = queue.popleft()
            yield node
            queue.extend(node.children)

    # ---------------
    # Tree properties
    # ---------------
    # Properties and methods related to tree metrics like leaf count, depth, height, etc.

    @property
    def n_leaves(self) -> int:
        return len(self.get_leaves())

    def branch_length_or_raise(self) -> float:
        if self.parent is None:
            return 0 if self.branch_length is None else self.branch_length
        if self.branch_length is None:
            raise ValueError(f"Branch length of node {self.name} is not set.")
        return self.branch_length

    @property
    def depth_level(self) -> int:
        return 0 if self.parent is None else self.parent.depth_level + 1

    @property
    def depth(self) -> float:
        parent_depth = 0 if self.parent is None else self.parent.depth
        return parent_depth + self.branch_length_or_raise()

    @property
    def height_level(self) -> int:
        if self.is_leaf():
            return 0
        return 1 + max(child.height_level for child in self.children)

    @property
    def height(self) -> float:
        if self.is_leaf():
            return 0.0
        return max(
            child.branch_length_or_raise() + child.height for child in self.children
        )

    # -------------
    # Miscellaneous
    # -------------
    # Other useful miscellaneous methods.

    def ladderize(self, key: Callable[["Tree"], Any] | None = None) -> None:
        if key is None:
            key = lambda node: node.n_leaves
        self._children.sort(key=key)
        for child in self.children:
            child.ladderize(key)

    def get_node(self, name: str) -> "Tree":
        for node in self:
            if node.name == name:
                return node
        raise ValueError(f"Node with name {name} not found.")

    def copy(self):
        new_tree = Tree(self.name, self.branch_length)
        new_tree.update(self.metadata)
        for child in self.children:
            new_tree.add_child(child.copy())
        return new_tree

    # ----------------
    # Dunder methods
    # ----------------
    # Special methods for standard behaviors like iteration, length, and representation.

    def __iter__(self) -> Iterator["Tree"]:
        return self.preorder_traversal()

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __repr__(self) -> str:
        return f"TreeNode(name='{self.name}', branch_length={self.branch_length}, metadata={self.metadata})"



from enum import Enum
from typing import Any, Callable

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1.inset_locator import inset_axes  # pyright: ignore

from phylogenie.treesimulator import Tree, get_node_depth_levels, get_node_depths


class Coloring(str, Enum):
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"


Color = str | tuple[float, float, float] | tuple[float, float, float, float]


def _draw_colored_tree(tree: Tree, ax: Axes, colors: Color | dict[Tree, Color]) -> Axes:
    if not isinstance(colors, dict):
        colors = {node: colors for node in tree}

    xs = (
        get_node_depth_levels(tree)
        if any(node.branch_length is None for node in tree.iter_descendants())
        else get_node_depths(tree)
    )
    ys: dict[Tree, float] = {node: i for i, node in enumerate(tree.get_leaves())}
    for node in tree.postorder_traversal():
        if node.is_internal():
            ys[node] = sum(ys[child] for child in node.children) / len(node.children)

    for node in tree:
        x1, y1 = xs[node], ys[node]
        if node.parent is None:
            ax.hlines(y=y1, xmin=0, xmax=x1, color=colors[node])  # pyright: ignore
            continue
        x0, y0 = xs[node.parent], ys[node.parent]
        ax.vlines(x=x0, ymin=y0, ymax=y1, color=colors[node])  # pyright: ignore
        ax.hlines(y=y1, xmin=x0, xmax=x1, color=colors[node])  # pyright: ignore

    ax.set_yticks([])  # pyright: ignore
    return ax


def draw_tree(
    tree: Tree,
    ax: Axes | None = None,
    color_by: str | dict[Tree, Any] | None = None,
    coloring: str | Coloring | None = None,
    default_color: Color = "black",
    cmap: str | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    show_legend: bool = True,
    labels: dict[Any, Any] | None = None,
    legend_kwargs: dict[str, Any] | None = None,
    show_hist: bool = True,
    hist_kwargs: dict[str, Any] | None = None,
    hist_axes_kwargs: dict[str, Any] | None = None,
) -> Axes | tuple[Axes, Axes]:
    if ax is None:
        ax = plt.gca()

    if color_by is None:
        return _draw_colored_tree(tree, ax, colors=default_color)

    if isinstance(color_by, dict):
        features = {node: color_by[node] for node in tree if node in color_by}
    else:
        features = {node: node[color_by] for node in tree if color_by in node.metadata}

    if coloring is None:
        coloring = (
            Coloring.CONTINUOUS
            if any(isinstance(f, float) for f in features.values())
            else Coloring.DISCRETE
        )

    def _get_colors(feature_map: Callable[[Any], Color]) -> dict[Tree, Color]:
        return {
            node: feature_map(features[node]) if node in features else default_color
            for node in tree
        }

    if coloring == Coloring.DISCRETE:
        if any(isinstance(f, float) for f in features.values()):
            raise ValueError(
                "Discrete coloring selected but feature values are not all categorical."
            )

        colormap = plt.get_cmap("tab20" if cmap is None else cmap)
        feature_colors = {
            f: mcolors.to_hex(colormap(i)) for i, f in enumerate(set(features.values()))
        }
        colors = _get_colors(lambda f: feature_colors[f])

        if show_legend:
            legend_handles = [
                mpatches.Patch(
                    color=feature_colors[f],
                    label=str(f) if labels is None else labels[f],
                )
                for f in feature_colors
            ]
            if any(color_by not in node.metadata for node in tree):
                legend_handles.append(mpatches.Patch(color=default_color, label="NA"))
            if legend_kwargs is None:
                legend_kwargs = {}
            ax.legend(handles=legend_handles, **legend_kwargs)  # pyright: ignore

        return _draw_colored_tree(tree, ax, colors)

    if coloring == Coloring.CONTINUOUS:
        vmin = min(features.values()) if vmin is None else vmin
        vmax = max(features.values()) if vmax is None else vmax
        norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
        colormap = plt.get_cmap("viridis" if cmap is None else cmap)
        colors = _get_colors(lambda f: colormap(norm(float(f))))

        if show_hist:
            default_hist_axes_kwargs = {"width": "25%", "height": "25%"}
            if hist_axes_kwargs is not None:
                default_hist_axes_kwargs.update(hist_axes_kwargs)
            hist_ax = inset_axes(ax, **default_hist_axes_kwargs)  # pyright: ignore

            hist_kwargs = {} if hist_kwargs is None else hist_kwargs
            _, bins, patches = hist_ax.hist(features, **hist_kwargs)  # pyright: ignore

            for patch, b0, b1 in zip(  # pyright: ignore
                patches, bins[:-1], bins[1:]  # pyright: ignore
            ):
                midpoint = (b0 + b1) / 2  # pyright: ignore
                patch.set_facecolor(colormap(norm(midpoint)))  # pyright: ignore
            return _draw_colored_tree(tree, ax, colors), hist_ax  # pyright: ignore

        else:
            sm = plt.cm.ScalarMappable(cmap=colormap, norm=norm)
            ax.get_figure().colorbar(sm, ax=ax)  # pyright: ignore
            return _draw_colored_tree(tree, ax, colors)

    raise ValueError(
        f"Unknown coloring method: {coloring}. Choices are {list(Coloring)}."
    )
